import os

try:
    from google.colab import auth
    from oauth2client.client import GoogleCredentials
    from pydrive.auth import GoogleAuth
    from pydrive.drive import GoogleDrive
except:
    pass
from pyunpack import Archive


def authenticate_colab():
    """
    Ask the user to connect to his google account to access google drive and google storage.

    Returns:
        google drive object
    """
    auth.authenticate_user()
    gauth = GoogleAuth()
    gauth.credentials = GoogleCredentials.get_application_default()
    drive = GoogleDrive(gauth)
    return drive


def download_dataset(file_id, save_path, drive):
    """
    Download compressed dataset (zip or 7z format) from google drive and extract it.

    Arguments:
        file_id (str): id of the file to download
        save_path (str): location where data is extracted
        drive (google drive object): return of authenticate_colab function
    """
    file = drive.CreateFile({'id': file_id})
    filename = file['title']
    print('Downloading {}'.format(filename))
    file.GetContentFile(filename)
    print('File {} downloaded'.format(filename))
    print('Extracting images')
    try:
        Archive(filename).extractall(save_path)
        print('Images extracted')
        os.remove(filename)
    except Exception as e:
        print('Could not extract images because of {}'.format(e))



import os
import sys

import tensorflow.keras.backend as K
from tensorflow.keras.losses import categorical_crossentropy as logloss
from tensorflow.keras.losses import sparse_categorical_crossentropy as sparselogloss

try:
    from decavision.utils.colab_utils import authenticate_colab
except:
    pass


class CheckpointDownloader(object):
    """
    Class to download current state to Google Drive after each iteration of hyperparameter optimization.
    To be used as callback for scikit-optimize routine. Files are saved in a folder called Checkpoints.

    Arguments:
        checkpoint_path (str): location where the checkpoint files are;
    """

    def __init__(self, checkpoint_path):
        self.checkpoint_path = checkpoint_path

    def __call__(self, res):
        """
        If working on colab, upload checkpoint file to google drive.

        Arguments:
            res (scipy object): The optimization as a OptimizeResult object.
        """
        if 'google.colab' in sys.modules and os.path.exists(self.checkpoint_path):
            print('Uploading checkpoint ' + self.checkpoint_path + ' to Google Drive')
            drive = authenticate_colab()
            file_id = None
            # get the id of the Checkpoints folder and create it if it doesn't exist already
            file_list = drive.ListFile({'q': "'root' in parents and trashed=false"}).GetList()
            for file in file_list:
                if file['title'] == "Checkpoints":
                    file_id = file['id']
            if not file_id:
                folder_metadata = {'title': 'Checkpoints', 'mimeType': 'application/vnd.google-apps.folder'}
                folder = drive.CreateFile(folder_metadata)
                folder.Upload()
                file_id = folder['id']

            file1 = drive.CreateFile({"title": "checkpoint.pkl",
                                      "parents": [{"kind": "drive#fileLink", "id": file_id}]})
            file1.SetContentFile(self.checkpoint_path)
            file1.Upload()


def custom_loss(old_logits, new_logits, old_classes, L=5, temp=5):
    """
    Distilling loss used when updating a model with new classes.
    It forces the model to remember what it learned about the old classes.
    Increasing the parameter L forces the model to remember more.
    High temperature puts more importance on the dominant classes and low temperature focuses on everything.

    Arguments:
        old_logits (keras tensor): classification layer of the old model, without the activation
        new_logits (keras tensor): classification layer of the new model, without the activation
        old_classes (int): number of classes in the old model
        L (float): parameter that controls how much information to remember
        temp (float): parameter that controls how many classes are important

    Returns:
        keras loss: loss function to be used during training
    """
    def loss(y_true, y_pred):
        y_soft = K.softmax(old_logits / temp)

        logits_pred = new_logits[:, :old_classes]
        y_pred_soft = K.softmax(logits_pred / temp)

        return sparselogloss(y_true, y_pred) + L * logloss(y_soft, y_pred_soft)

    return loss


def f1_score(y_true, y_pred):
    """
    Computer F1-score metric

    Arguments:
        y_true (tensor): True labels
        y_pred (tensor): Predicted labels
    """
    # recall
    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    possible_positives = K.sum(K.round(K.clip(y_true, 0, 1)))
    recall = true_positives / (possible_positives + K.epsilon())

    # precision
    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    predicted_positives = K.sum(K.round(K.clip(y_pred, 0, 1)))
    precision = true_positives / (predicted_positives + K.epsilon())

    return 2 * ((precision * recall) / (precision + recall + K.epsilon()))



import os
import math
import shutil

import Augmentor


class DataAugmentor:
    """
    Class to generate augmented images for classification purposes. Images must be located in a folder with
    subfolders for each category.

    Arguments:
        path (str): location of the folders containing images for each category
        option (bool): use or not option during the augmentation, option can be anything between distortion,
            flip_horizontal, flip_vertical, random_crop, random_erasing, rotate, resize, skew, shear, brightness,
            contrast and color (options described at https://github.com/mdbloice/Augmentor)
        target_width (int): new width of resized images (if resized is True)
        target_height (int): new height of resized images (if resized is True)
    """
    
    def __init__(self, path, distortion=False, flip_horizontal=False,
                 flip_vertical=False, random_crop=False, random_erasing=False,
                 rotate=False, resize=False, skew=False,
                 shear=False, brightness=False, contrast=False,
                 color=False, target_width=299, target_height=299):
        self.path = path
        self.categories = sorted(os.listdir(self.path))
        self.categories_folder = [os.path.abspath(os.path.join(self.path, i)) for i in self.categories]
        # options
        self.distortion = distortion
        self.flip_horizontal = flip_horizontal
        self.flip_vertical = flip_vertical
        self.random_crop = random_crop
        self.random_erasing = random_erasing
        self.rotate = rotate
        self.resize = resize
        self.skew = skew
        self.shear = shear
        self.brightness = brightness
        self.contrast = contrast
        self.color = color
        self.target_width = target_width
        self.target_height = target_height
        # Create a pipeline for each class
        self.pipelines = {}
        for folder in self.categories_folder:
            print("Folder %s:" % folder)
            self.pipelines[os.path.split(folder)[1]] = (Augmentor.Pipeline(folder))
            print("\n----------------------------\n")
            
        for p in self.pipelines.values():
            print("Class %s has %s samples." % (p.augmentor_images[0].class_label, len(p.augmentor_images)))
    
    def _set_options(self, pipeline):
        """
        Set desired options to the augmentation pipeline.

        Arguments:
            pipeline (obj): augmentor object to do data augmentation
        """
        if self.distortion: 
            pipeline.random_distortion(probability=1, grid_width=4, grid_height=4, magnitude=8)
        if self.flip_horizontal: 
            pipeline.flip_left_right(probability=0.5)
        if self.flip_vertical: 
            pipeline.flip_top_bottom(probability=0.5)
        if self.random_crop: 
            pipeline.crop_random(probability=0.5, percentage_area=0.75)
        if self.resize: 
            pipeline.resize(probability=1, width=self.target_width, height=self.target_height, resample_filter="BILINEAR")
        if self.random_erasing: 
            pipeline.random_erasing(probability=0.5, rectangle_area=0.25)
        if self.rotate: 
            pipeline.rotate(0.5, max_left_rotation=10, max_right_rotation=10)
        if self.skew: 
            pipeline.skew(probability=0.5)
        if self.shear: 
            pipeline.shear(probability=0.25, max_shear_left=10, max_shear_right=10)
        if self.brightness: 
            pipeline.random_brightness(probability=0.25, min_factor=0.7, max_factor=1.3)
        if self.contrast: 
            pipeline.random_contrast(probability=0.25, min_factor=0.7, max_factor=1.3)
        if self.color: 
            pipeline.random_color(probability=0.25, min_factor=0.7, max_factor=1.3)
        
    def generate_images_single_class(self, class_size_approximation, category):
        """
        Generate augmented images for a single category according to the specified options. Images are generated
        until the desired size is reached and they are saved in an outputs folder at the location of the original
        images.

        Arguments:
            class_size_approximation (int): approximate total number of images wanted
            category (str): folder name where images to augment are located in the main folder
        """
        pipeline = self.pipelines[category]
        self._set_options(pipeline)
        for _ in range(math.ceil(class_size_approximation/len(pipeline.augmentor_images)-1)):
            pipeline.process()
        self._move_outputs(pipeline)
        self._delete_outputs(pipeline)
    
    def generate_images(self, class_size_approximation):
        """
        Generate augmented images for all categories in the main folder according to the specified options. Images
        are generated until the desired size is reached and they are saved in the same subfolder as the original
        images.

        Arguments:
            class_size_approximation (int): approximate total number of images wanted for each category
        """
        for pipeline in self.pipelines.values():
            self._set_options(pipeline)
            for _ in range(math.ceil(class_size_approximation/len(pipeline.augmentor_images)-1)):
                pipeline.process()
            self._move_outputs(pipeline)
            self._delete_outputs(pipeline)

    def _move_outputs(self, pipeline):
        """
        Move output folder content to main folder.

        Arguments:
            pipeline (obj): augmentor object to do data augmentation
        """
        output_dir = pipeline.augmentor_images[0].output_directory
        class_label = pipeline.augmentor_images[0].class_label

        for image in os.listdir(pipeline.augmentor_images[0].output_directory):
            os.rename(os.path.join(output_dir, image), os.path.join(os.path.join(self.path, class_label), image))

    def _delete_outputs(self, pipeline):
        """
        Delete output folder if present.

        Arguments:
            pipeline (obj): augmentor object to do data augmentation
        """
        dirpath = pipeline.augmentor_images[0].output_directory
        if os.path.exists(dirpath) and os.path.isdir(dirpath):
            shutil.rmtree(dirpath)



import os
from random import shuffle
import sys
import tarfile
import urllib.request
import zipfile

import numpy as np
import PIL
from PIL import Image
import tensorflow as tf


def prepare_image(image_path, target_size, rescaling=255):
    """
    Load and convert image to numpy array to feed it to a neural network. Image is resized, converted to RGB
    and its pixels are normalized if required. An extra dimension is added to the array.

    Arguments:
        image_path (str): path to image to be converted
        target_size (tuple(int,int)): desired size for the image
        rescaling (int): divide all the pixels of the image by this number

    Returns:
        numpy array: processed image, with shape (1,target_size,3)
    """
    image = Image.open(image_path)
    # reshape the image
    image = image.resize(target_size, PIL.Image.BILINEAR).convert("RGB")
    # convert the image into a numpy array, and expend to a size 4 tensor
    image = tf.keras.preprocessing.image.img_to_array(image)
    image = np.expand_dims(image, axis=0)
    # rescale the pixels to a 0-1 range
    image = image.astype(np.float32) / rescaling
    return image


def check_RGB(path, target_size=None):
    """
    Convert all images in a folder into RGB format and resize them if desired. Images that
    can't be opened are deleted. Folder must contain a subfolder for each category.

    Arguments:
        path (str): path to the image directory
        target_size (tuple(int,int)): if specified, images are resized to this size
    """

    classes = os.listdir(path)
    classes_paths = [os.path.abspath(os.path.join(path, i)) for i in classes]

    counter = 0
    for i in classes_paths:
        imgs = os.listdir(i)
        imgs_paths = [os.path.abspath(os.path.join(i, j)) for j in imgs]
        # Loop through all the images in the path
        for img in imgs_paths:
            # try to open it
            try:
                if target_size is not None:
                    jpg = Image.open(img).resize(target_size, PIL.Image.BILINEAR).convert('RGB')
                else:
                    jpg = Image.open(img).convert('RGB')
                jpg.save(str(img))
            except:
                # delete the file
                print('Deleting', img)
                os.remove(img)
            counter += 1
            if counter % 1000 == 1:
                print('Verified', counter, 'images')


def create_dir(path):
    """
    Check if directory exists and create it if it does not.

    Arguments:
        path (str): path to directory to create
    """
    if not os.path.exists(path):
        os.mkdir(path)


def split_train(path='data/image_dataset',
                split=0.1,
                multilabel=False,
                with_test=False):
    """
    Separate images randomly into a training, a validation and potentially a test dataset.
    Images must be located in a folder called train, which contains a subfolder per category.
    Val and potentially test folders will be created amd images moved into it from train.

    Arguments:
        path (str): path to the image_dataset directory
        split (float): fraction of each category that we move to the validation (val) subdirectory
        multilabel (bool): if we are in a multilabel setting, as such not labels in the image names
        with_test (bool): determine if one image of each category is moved to test dataset
    """

    # Create a val subdirectory
    create_dir(path + '/val')
    # Create a test subdirectory
    if with_test:
        create_dir(path + '/test')

    # Loop through all the categories in the train directory
    if multilabel:
        list_labels = ['']
    else:
        list_labels = os.listdir(path + '/train')
    for i in list_labels:

        # Create the folder in the val subdirectory
        create_dir(path + '/val/' + i)

        # extract and shuffle all the images
        images = os.listdir(path + '/train/' + i)
        shuffle(images)

        # Move a fraction of the images to the val directory
        for j in range(int(split * len(images))):
            if i != '':
                os.rename(path + '/train/' + i + '/' + images[j], path + '/val/' + i + '/' + images[j])
            else:
                os.rename(path + '/train/' + images[j], path + '/val/' + images[j])

        # Move one of the images to the test directory
        if with_test:
            # Create the folder in the val subdirectory
            create_dir(path + '/test/' + i)

            for j in range(int(split * len(images)), 2 * int(split * len(images))):
                if i != '':
                    os.rename(path + '/train/' + i + '/' + images[j], path + '/test/' + i + '/' + images[j])
                else:
                    os.rename(path + '/train/' + images[j], path + '/val/' + images[j])
    print('Training dataset has been split.')


def print_download_progress(count, block_size, total_size):
    """
    Function used for printing the download progress. Inspired by:
    https://github.com/Hvass-Labs/TensorFlow-Tutorials/blob/master/download.py
    """
    # Percentage completion.
    pct_complete = float(count * block_size) / total_size

    # Limit it because rounding errors may cause it to exceed 100%.
    pct_complete = min(1.0, pct_complete)

    # Status-message. Note the \r which means the line should overwrite itself.
    msg = "\r- Download progress: {0:.1%}".format(pct_complete)

    # Print it.
    sys.stdout.write(msg)
    sys.stdout.flush()


def download_dataset(download_dir='data/',
                     url='http://data.csail.mit.edu/places/places365/places365standard_easyformat.tar'):
    """
    Download a dataset in format .zip, .tar, .tar.gz or .tgz and extract the data.
    Inspired by: https://github.com/Hvass-Labs/TensorFlow-Tutorials/blob/master/download.py

    Arguments:
        download_dir (str): folder to store the data
        url (str): location of the dataset on the internet
    """
    # Filename for saving the file downloaded from the internet.
    # Use the filename from the URL and add it to the download_dir.
    filename = url.split('/')[-1]
    file_path = os.path.join(download_dir, filename)

    # Check if the file already exists.
    # If it exists then we assume it has also been extracted,
    # otherwise we need to download and extract it now.
    if not os.path.exists(file_path):
        # Check if the download directory exists, otherwise create it.
        if not os.path.exists(download_dir):
            os.makedirs(download_dir)

        # Download the file from the internet.
        file_path, _ = urllib.request.urlretrieve(url=url,
                                                  filename=file_path,
                                                  reporthook=print_download_progress)

        print("\n Download finished. Extracting files. \n")

        if file_path.endswith(".zip"):
            # Unpack the zip-file.
            with zipfile.ZipFile(file=file_path, mode="r") as f:
                f.extractall(download_dir)

        elif file_path.endswith((".tar.gz", ".tgz")):
            # Unpack the tar-ball.
            with tarfile.open(name=file_path, mode="r:gz") as f:
                f.extractall(download_dir)

        elif file_path.endswith(".tar"):
            # Unpack tar file.
            with tarfile.open(name=file_path, mode="r") as f:
                f.extractall(download_dir)

        print("Done.")
    else:
        print("Data has apparently already been downloaded and unpacked.")



import ast
import os
import shutil
from distutils.dir_util import copy_tree

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.keras.models import load_model
from tqdm import tqdm


class PseudoLabelGenerator:
    """
    Class to generate pseudo labels for unlabeled images using a trained model.

    Arguments:
        model_path (str): location of the h5 tensorflow model to use
        train_data_path (str): folder which holds training data
        unlabeled_path (str): folder which holds unlabeled data
        pseudo_data_path (str): folder to store training data and pseudo data combined
        output_folder (str): folder to store outputs
        csv_filename (str): name of csv file
    """

    def __init__(self, model_path="model.h5", train_data_path="data/image_dataset/train",
                 unlabeled_path="data/unlabeled", pseudo_data_path="data/train_ssl",
                 output_folder="outputs", csv_filename="data.csv"):

        self.train_data_path = train_data_path
        self.unlabeled_path = unlabeled_path
        self.pseudo_data_path = pseudo_data_path
        self.output_folder = output_folder
        self.csv_path = os.path.join(self.output_folder, csv_filename)

        # Load model
        self.model = load_model(model_path, compile=False)
        print("Loaded model.")

        # Make new output folder
        if not os.path.exists(self.output_folder):
            os.mkdir(self.output_folder)

        # Make dictionary for classes and their index
        self.class_names = sorted(os.listdir(self.train_data_path))
        self.class_dict = {cat: i for (i, cat) in enumerate(self.class_names)}

    def _load_img(self, path, target_size=(299, 299)):
        """
        Load an image from a given path and normalize it

        Arguments:
            path (list): Input image path
            target_size (tuple): Size of image
        Returns:
            np.array: Numpy array of the data
        """
        # Read image
        bits = tf.io.read_file(path)
        image = tf.image.decode_jpeg(bits, channels=3)
        # Resize
        image = tf.image.resize(image, size=[*target_size])
        image = tf.reshape(image, [*target_size, 3])
        image = tf.cast(image, tf.uint8)
        # Normalize [0, 1]
        image = tf.image.convert_image_dtype(image, dtype=tf.float32)
        image = image.numpy()
        return image

    def _plot_data(self, predictions, name, output_path):
        """
        Plots a bar plot and saves it to a file.

        Arguments:
            predictions (list): List of predictions
            name (str): Title of the plot
            output_path (str): Save file using this name
        """
        predictions = sorted(predictions)
        samples = list(range(len(predictions)))
        plt.bar(samples, predictions, color='g')
        plt.axhline(y=0.5, color='r', linestyle='--')
        plt.title(name, size=16)
        plt.xlabel("Number of unlabelled images", size=16)
        plt.ylim([0.0, 1.0])
        plt.ylabel("Probability", size=16)
        plt.tick_params(labelright=True)
        plt.savefig(output_path, dpi=100)
        plt.clf()  # clear buffer, otherwise plot overlap!

    def plot_confidence_scores(self, per_class=True, overall=True):
        """
        Generate bar plots for highest confidence predictions per class and overall and save them.

        Arguments:
            per_class (bool): make bar plots per class or not
            overall (bool): make overall bar plot or not
        """
        dt = pd.read_csv(self.csv_path)
        dt['All Class Predictions List'] = dt['All Class Predictions'].apply(
            lambda x: ast.literal_eval(x))

        raw_predictions_ = dt[["Highest Confidence"]].values
        raw_predictions = [pred[0] for pred in raw_predictions_]

        raw_predictions_all_ = dt[['All Class Predictions List']].values
        raw_predictions_all = [pred[0] for pred in raw_predictions_all_]

        # Plot graph for highest confidence pseudo labels for each class
        if per_class:
            for idx, cat in enumerate(self.class_names):
                predictions = [pred[idx] for pred in raw_predictions_all]
                title = "Confidences for the class: {}".format(cat)
                path = "{}/{}_confidences.png".format(
                    self.output_folder, cat)
                self._plot_data(predictions, title, path)

        # Plot graph for highest confidence pseudo labels for all unlabeled images
        if overall:
            self._plot_data(raw_predictions,
                            name="Highest confidence pseudo labels",
                            output_path="{}/highest_confidence_predictions.png".format(
                                self.output_folder))

    def _make_dataset(self, filenames, batch_size):
        def parse_image(filename):
            image = tf.io.read_file(filename)
            image = tf.image.decode_jpeg(image, channels=3)
            image = tf.image.resize(image, [299, 299])
            image = tf.cast(image, tf.uint8)
            image = tf.image.convert_image_dtype(image, dtype=tf.float32)
            return image

        def configure_for_performance(ds):
            ds = ds.batch(batch_size)
            ds = ds.prefetch(buffer_size=tf.data.experimental.AUTOTUNE)
            return ds

        filenames_ds = tf.data.Dataset.from_tensor_slices(filenames)
        images_ds = filenames_ds.map(parse_image,
                                     num_parallel_calls=tf.data.experimental.AUTOTUNE)
        ds = configure_for_performance(images_ds)
        return ds

    def move_unlabeled_images(self, threshold=None):
        """
        Split unlabeled images into folders of the training data based on pseudo labels. A new training dataset is
        created with the labeled and pseudo labeled data.

        Arguments:
            threshold (float): Discard images with prediction below this confidence, default is None.
        """

        # Copy the training/labeled data to the destination folder where
        # we will also store the pseudo labels.
        copy_tree(self.train_data_path, self.pseudo_data_path)

        dt = pd.read_csv(self.csv_path)
        filepaths = dt[["Filepaths"]].values
        predicted_class = dt[["Predicted Class"]].values
        raw_predictions = dt[["Highest Confidence"]].values

        for proba, y, path in zip(raw_predictions, predicted_class, filepaths):
            # The results variable should be the same as the class category
            for class_name, index in self.class_dict.items():
                if threshold:
                    # For thresholding predictions
                    if index == y and proba >= threshold:
                        shutil.copy(os.path.join(self.unlabeled_path, str(path[0])),
                                    os.path.join(self.pseudo_data_path, class_name))
                else:
                    # For hard predictions
                    if index == y:
                        shutil.copy(os.path.join(self.unlabeled_path, str(path[0])),
                                    os.path.join(self.pseudo_data_path, class_name))
        print("Moved unlabeled images to their pseudo label categories.")

    def generate_pseudolabel_data(self, plot_confidences=False, threshold=None, move_images=False, batch_size=32):
        """ Use trained model to make pseudo labels and save them into a csv file. Also possible to plot the results
            and move the unlabeled images directly to the category corresponding to their pseudo label.

            Arguments:
                plot_confidences (boolean): Whether to plot confidence graphs for raw confidences and per class confidences.
                threshold (float): Discard images with prediction below this confidence, default is None. Only used
                    if move_images is True.
                move_images (bool): Move images into categories or not
                batch_size (int): Batch size while making predictions

            Returns:
                pseudo_data_path: A folder with both labeled and pseudo labeled images.
        """

        print("Generating pseudo labels...")
        # Generate pseudo labels
        unlabeled_image_paths = os.listdir(self.unlabeled_path)
        print("There are {} unlabeled images.".format(
            len(unlabeled_image_paths)))

        raw_predictions = []  # single confidence value of predicted class
        predicted_class = []  # predicted class index
        raw_predictions_all = []  # confidences for all classes

        unlabeled_filenames = [os.path.join(self.unlabeled_path,
                                            path) for path in unlabeled_image_paths]
        ds = self._make_dataset(unlabeled_filenames, batch_size)

        y_preds = []
        for batch in tqdm(ds):
            y_preds_ = self.model.predict(batch)
            y_preds.extend(list(y_preds_))

        for y_pred in y_preds:
            y = np.argmax(y_pred)
            # Get probability score
            proba = y_pred[y]
            predicted_class.append(y)
            raw_predictions.append(proba)
            raw_predictions_all.append(list(y_pred))
        raw_predictions_paths = [path for path in unlabeled_image_paths]

        # 'Pseudo Class Names': pseudo_class_names,
        print("Saving CSV with pseudo predictions.")
        data = {'Filepaths': raw_predictions_paths,
                'Predicted Class': predicted_class,
                'Highest Confidence': raw_predictions,
                'All Class Predictions': raw_predictions_all}
        df = pd.DataFrame(data)
        df.to_csv(self.csv_path, index=False)

        # move pseudo labeled images
        if move_images:
            self.move_unlabeled_images(threshold=threshold)

        if plot_confidences:
            print("Plotting data.")
            self.plot_confidence_scores()



import numpy as np
import tensorflow as tf

from decavision.model_training.tfrecords_image_classifier import ImageClassifier
from decavision.utils import training_utils
from decavision.utils import utils


AUTO = tf.data.experimental.AUTOTUNE


class ProgressiveLearner(ImageClassifier):
    """
    Class to update an already trained model with new classes without losing too
    much of the information learned about the old classes.

    Arguments:
        tfrecords_folder (str): location of tfrecords (can be on google storage if authenticated), saved in
            folders train and val, filenames of the form filenumber-numberofimages.tfrec
        model_path (str): path to .h5 model trained with this library on the old classes
        transfer_model (str): pretrained model that was used to train the old model, can be one of Inception,
            Xception, Inception_Resnet, Resnet, B0, B3, B5, B7, V2-S, V2-M, V2-L or V2-XL
        batch_size (int): size of batches of data used for training
    """

    def __init__(self, tfrecords_folder, model_path, transfer_model, batch_size=128):
        super().__init__(tfrecords_folder, batch_size, transfer_model)
        self.distil = False
        self.temp = 5
        self.L = 5
        self.old_model = utils.load_model_clear(model_path)
        self.logits = self.old_model.layers[-2].output

    def _update_model(self):
        """
        Take the old model, remove the classification layer and add a new one that includes
        the new classes. The weights connecting to the old classes are kept.

        Returns:
            tf.keras model: model with updated classification layer
        """
        self.old_classes = self.old_model.layers[-1].output_shape[-1]
        self.updated_classes = len(self.categories)
        self.new_classes = self.updated_classes - self.old_classes
        print('Updating output layer from {} classes to {} classes'.format(self.old_classes, self.updated_classes))

        # get weights connecting last layer to output and initialize new weights using them
        weights = self.old_model.get_weights()[-2]
        weights = np.append(weights, np.random.uniform(-1, 1, size=(weights.shape[0], self.new_classes)), axis=1)
        bias = self.old_model.get_weights()[-1]
        bias = np.append(bias, np.zeros((self.new_classes,)))

        # define new model without the old ouputs and add new output
        updated_logits = tf.keras.layers.Dense(self.updated_classes, weights=(weights, bias), name='updated_logs')(
            self.old_model.layers[-3].output)
        updated_predictions = tf.keras.layers.Activation('softmax', name='updated_preds')(updated_logits)

        return tf.keras.models.Model(inputs=self.old_model.inputs, outputs=updated_predictions)

    def _create_model(self, *args, **kwargs):
        """
        Create a keras model from a model with fewer classes.

        Returns:
            tf.keras model: pretrained model with added classes
            int: index of the layer where the last block of the pretrained model starts
            str: name of the loss function
            list(str): names of the metrics to use
        """
        print('Creating model')
        # get the last block for the fine tuning process
        if self.transfer_model == 'Xception':
            base_model_last_block = 116  # last block 126, two blocks 116
        elif self.transfer_model == 'Inception_Resnet':
            base_model_last_block = 287  # last block 630, two blocks 287
        elif self.transfer_model == 'Resnet':
            base_model_last_block = 155  # last block 165, two blocks 155
        elif self.transfer_model == 'B0':
            base_model_last_block = 213  # last block 229, two blocks 213
        elif self.transfer_model == 'B3':
            base_model_last_block = 354  # last block 370, two blocks 354
        elif self.transfer_model == 'B5':
            base_model_last_block = 417  # last block 559, two blocks 417
        elif self.transfer_model == 'B7':
            base_model_last_block = None  # all layers trainable
        elif self.transfer_model in ['V2-S', 'V2-M', 'V2-L', 'V2-XL']:
            base_model_last_block = None  # all layers trainable
        else:
            base_model_last_block = 249  # last block 280, two blocks 249

        self.model = self._update_model()
        self.updated_logits = self.model.layers[-2].output

        # Define the optimizer and the loss and the optimizer
        loss = 'sparse_categorical_crossentropy'
        if self.distil and not self.use_TPU:
            loss = training_utils.custom_loss(self.logits, self.updated_logits, self.old_classes, self.temp, self.L)
        metrics = ['sparse_categorical_accuracy']

        return self.model, base_model_last_block, loss, metrics

    def fit(self, learning_rate=1e-3, learning_rate_fine_tuning=1e-4,
            epochs=5, save_model=False, verbose=True,
            fine_tuning=True, logs=None):
        """
        Train an image classification model based on a model trained with a smaller number of classes.
        The whole model is trained, unless there is some fine tuning, in which case a second round of
        training is done with the last layers of the model unfrozen. Training can be stopped if
        no sufficient improvement in accuracy.

        Arguments:
            learning_rate (float): learning rate used when training whole model
            learning_rate_fine_tuning (float): learning rate used when fine tuning last layers
            epochs (int): number of epochs done when training (doubled if fine tuning)
            save_model (str): specify a name for the trained model to save it, model is saved in .h5 if
                the name contains the extension and in .pb if no extension in the name
            verbose (bool): show details of training or not
            fine_tuning (bool): fine tune last layers or not
            min_accuracy (float): if specified, stop training when improvement in accuracy is smaller than min_accuracy
            logs (str): if specified, tensorboard is used and logs are saved at this location
        """
        super().fit(learning_rate=learning_rate, learning_rate_fine_tuning=learning_rate_fine_tuning,
                    epochs=epochs, save_model=save_model, verbose=verbose,
                    fine_tuning=fine_tuning, logs=logs)

    def hyperparameter_optimization(self):
        """
        This class is not implemented for progressive learning.
        """
        pass



import csv
import os
import sys

from google.cloud import storage
import tensorflow as tf


def load_model_clear(path, include_top=True):
    """
    Clear tensorflow session and load keras .h5 model.

    Arguments:
        path (str): location of the model, including its name
        include_top (bool): whether or not to include the last layer of the model

    Returns:
        tf.keras model: loaded model
    """
    tf.keras.backend.clear_session()
    model = tf.keras.models.load_model(path, compile=include_top)
    print('Model loaded !')
    if not include_top:
        model = tf.keras.Model(inputs=model.input, outputs=model.layers[-2].output)
    return model


def check_PU():
    """
    Check if machine is running on TPU, GPU or CPU.

    Returns:
        bool: whether or not the machine runs on a TPU
        bool: whether or not the machine runs on a GPU
    """
    use_tpu = False
    use_gpu = False
    try:
        tpu = tf.distribute.cluster_resolver.TPUClusterResolver()  # TPU detection
        use_tpu = True
        print('Running on TPU ', tpu.cluster_spec().as_dict()['worker'])
    except ValueError:
        if tf.test.is_built_with_cuda():
            use_gpu = True
            print("Running on GPU", tf.test.gpu_device_name())
        else:
            print("Running on CPU")
    return use_tpu, use_gpu


def check_sep():
    """
    Check if the OS is windows or anything else to return the right separator.

    Returns:
         str: '\\' for windows and '/' for others.
    """
    if sys.platform == 'win32':
        return '\\'
    else:
        return '/'


def gcs_bucket(folder):
    """
    Create bucket object to be used to access files in google could storage bucket.

    Arguments:
        folder (str): name of the google storage folder, must be of the form gs://bucketname/prefix

    Returns:
        bucket object: google storage object
        string: prefix of the bucket, to be used to access files
    """
    path_split = folder.split("/")
    bucket_name = path_split[2]
    prefix = os.path.join(*path_split[3:])
    storage_client = storage.Client(project=None)
    bucket = storage_client.get_bucket(bucket_name)
    return bucket, prefix


def empty_folder(folder):
    """
    Delete all files in a given folder. First try to delete
    locally and if it fails try to delete in google cloud
    storage. If folder is in GCS, the link must include the
    gs:// part and the folder will be deleted as well.

    Arguments:
        folder (str): path of folder where files are located
    """
    if is_gcs(folder):
        bucket, prefix = gcs_bucket(folder)
        blobs = bucket.list_blobs(prefix=prefix)
        for blob in blobs:
            blob.delete()
        print("Deleted files in gcs directory", folder)
    else:
        for filename in os.listdir(folder):
            file_path = os.path.join(folder, filename)
            os.remove(file_path)
        print("Deleted files in local directory", folder)


def upload_file_gcs(gcp_path, file_path):
    """
    Copy local file to a folder in google cloud storage and call it classes.csv.

    Arguments:
        gcp_path (str): full path to gcp folder
        file_path (str): location of path to upload
    """
    bucket, prefix = gcs_bucket(gcp_path)
    blob = bucket.blob(prefix + '/classes.csv')
    blob.upload_from_filename(file_path)


def is_gcs(path):
    """
    Check if path is to a google cloud storage directory or a local one. Determined from the presence of 'gs'
    at the beginning of the path.

    Arguments:
        path (str): path to assess

    Returns:
         bool: True if path is on gcs and False if local
    """
    if path[:2] == "gs":
        return True
    else:
        return False


def load_classes(path, name='classes'):
    """
    Open csv and create list from its content. If csv is on google cloud storage,
    the file is downloaded into working directory and then opened.

    Arguments:
        path (str): location of the csv file
        name (str): name of csv file to open, without extension

    Returns:
        list: first line of the csv file"""
    if is_gcs(path):
        bucket, prefix = gcs_bucket(path)
        for blob in bucket.list_blobs(prefix=prefix + '/' + name + '.csv'):
            blob.download_to_filename(name + ".csv")
        with open('classes.csv', newline='') as f:
            reader = csv.reader(f)
            classes = next(reader)
    else:
        with open(os.path.join(path, name + '.csv'), newline='') as f:
            reader = csv.reader(f)
            classes = next(reader)
    return classes



import setuptools

with open('requirements.txt', 'r') as req:
    requirements = []
    for line in req:
        requirements.append(line.strip())

setuptools.setup(
    name="decavision",
    version="1.4.1",
    author="Décathlon Canada",
    author_email="sportvisionapi@decathlon.com",
    description="A package to easily train powerful image classification models using colab's free TPUs.",
    long_description="""The AI team at Décathlon Canada developed a library to help with the training of image
                        classification models. It is specially made to exploit the free TPUs that are offered
                        in Google colab notebooks. You can find the full documentation
                        [here](https://decavision-doc.herokuapp.com/)

                        ## Version 1.4.1
                        Fix hyperparameter optimization for multilabel

                        ## Version 1.4.0
                        Added multilabel classification

                        ## Version 1.3.0
                        Use suggested image sizes
                        Add EfficientNetV2 models

                        ## Version 1.2.1
                        Change to Tensorflow 2.5

                        ## Version 1.2.0
                        Add semi-supervised learning features

                        ## Version 1.1.3
                        Link to public repository

                        ## Version 1.1.2
                        Change name of package

                        ## Version 1.1.1
                        Fix typo in split_train

                        ## Version 1.1.0
                        Remove google scraping

                        ## Version 1.0.0
                        Original release""",
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=requirements,
    url='https://github.com/Decathlon/decavision.git'
)



import csv
import json
import math
import os

import tensorflow as tf

from decavision.utils import utils


class TfrecordsGenerator:
    """
    Class to transform images into tfrecords format to train neural networks. Resulting files can be
    saved to google storage or locally. Can't be used with a TPU because local files need to be read.
    Strongly inspired by: https://medium.com/@moritzkrger/speeding-up-keras-with-tfrecord-datasets-5464f9836c36
    """

    def __init__(self):
        pass

    def _to_tfrecord(self, img_bytes, label):
        """ Convert image and label to tfrecord example. """
        example = tf.train.Example(features=tf.train.Features(feature={
            'image': tf.train.Feature(bytes_list=tf.train.BytesList(value=[img_bytes])),
            'label': tf.train.Feature(int64_list=tf.train.Int64List(value=label))
        }))
        return example

    def convert_image_folder(self, img_folder='data/image_dataset/train',
                             output_folder='data/tfrecords_dataset/train',
                             multilabel=False,
                             img_folder_new=None,
                             target_size=None,
                             shards=16,
                             json_path=None):
        """
        Convert all images in a folder (like train or val) to tfrecords. Folder must contain subfolders for each category.
        Possibility to combine data from two folders to perform progressive learning. Tfrecords can be saved
        locally or on google storage. A csv file containing the names of the classes is also saved.

        For multilabel all the images must be in a single folder and there must exist a json file with the keys
        being the filenames and the values being lists of labels.

        Arguments:
            img_folder (str): location of the images
            output_folder (str): folder to save the results, content of folder is deleted to save new data
            img_folder_new (str): if specified, images from this folder are included in the tfrecords as
                new categories for the purpose of progressive learning
            multilabel (bool): True if it is a multilabel problem
            shards (int): number of files to create
            target_size (tuple(int,int)): size to reshape the images if desired
            json_path (str): location of the json file, only used for multilabel
        """
        # Create output directory if it does not exists
        if not os.path.exists(output_folder):
            if not utils.is_gcs(output_folder):
                os.makedirs(output_folder)
                print('Created directory {}'.format(output_folder))
        else:
            utils.empty_folder(output_folder)

        # Get all file names of images present in folder
        if multilabel:
            with open(json_path, 'rb') as f:
                labels_dict = json.load(f)
            classes = []
            for values in labels_dict.values():
                classes += values
            classes = sorted(list(set(classes)))
            img_pattern = os.path.join(img_folder, '*')
            nb_images = len(tf.io.gfile.glob(img_pattern))
            shard_size = math.ceil(1.0 * nb_images / shards)
        else:
            classes = sorted(os.listdir(img_folder))
            if img_folder_new:
                new_classes = sorted(os.listdir(img_folder_new))
                classes = classes + new_classes
            print(classes)
            img_pattern = os.path.join(img_folder, '*/*')
            if img_folder_new:
                img_pattern_new = os.path.join(img_folder_new, '*/*')
                img_pattern = [img_pattern, img_pattern_new]
            nb_images = len(tf.io.gfile.glob(img_pattern))
            shard_size = math.ceil(1.0 * nb_images / shards)
        print("Pattern matches {} images which will be rewritten as {} .tfrec files containing {} images each.".format(nb_images, shards, shard_size))

        def decode_jpeg_and_label(filename):
            bits = tf.io.read_file(filename)
            image = tf.image.decode_jpeg(bits, channels=3)
            label = tf.strings.split(tf.expand_dims(filename, axis=-1), sep=utils.check_sep())
            if multilabel:
                label = label.values[-1]
            else:
                label = label.values[-2]
            return image, label

        def resize_image(image, label):
            image = tf.image.resize(image, size=[*target_size])
            image = tf.reshape(image, [*target_size, 3])
            return image, label

        def recompress_image(image, label):
            image = tf.cast(image, tf.uint8)
            image = tf.image.encode_jpeg(image, quality=100, format='rgb',
                                         optimize_size=True, chroma_downsampling=False)
            return image, label

        AUTO = tf.data.experimental.AUTOTUNE

        filenames = tf.data.Dataset.list_files(img_pattern)  # This also shuffles the images
        dataset = filenames.map(decode_jpeg_and_label, num_parallel_calls=AUTO)
        if target_size:
            dataset = dataset.map(resize_image, num_parallel_calls=AUTO)
        dataset = dataset.map(recompress_image, num_parallel_calls=AUTO)
        dataset = dataset.batch(shard_size)  # sharding: there will be one "batch" of images per file

        print("Writing TFRecords")
        for shard_num, shard in enumerate(dataset):
            images, labels = shard
            images = images.numpy()
            labels = labels.numpy()
            # batch size used as shard size here
            shard_size = images.shape[0]
            # good practice to have the number of records in the filename
            filename = os.path.join(output_folder, "{:02d}-{}.tfrec".format(shard_num, shard_size))

            with tf.io.TFRecordWriter(filename) as out_file:
                for i in range(shard_size):
                    if multilabel:
                        label_list = [classes.index(x) for x in labels_dict[labels[i].decode('utf8')]]
                    else:
                        label_list = [classes.index(labels[i].decode('utf8'))]
                    example = self._to_tfrecord(images[i],  # re-compressed image: already a byte string
                                                label_list)
                    out_file.write(example.SerializeToString())
                print("Wrote file {} containing {} records".format(filename, shard_size))

        # save classes locally or on gcs depending on the output folder
        if utils.is_gcs(output_folder):
            with open("classes.csv", "w", newline="") as myfile:
                wr = csv.writer(myfile, quoting=csv.QUOTE_ALL)
                wr.writerow(classes)
            utils.upload_file_gcs(os.path.dirname(output_folder), "classes.csv")
            print("Saved classes to google storage")
        else:
            with open(os.path.join(os.path.dirname(output_folder), "classes.csv"), "w", newline="") as myfile:
                wr = csv.writer(myfile, quoting=csv.QUOTE_ALL)
                wr.writerow(classes)
            print("Saved classes locally")


if __name__ == '__main__':
    transformer = TfrecordsGenerator()



import glob
import math
import os
import random
import json
import cv2

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sn
import pandas as pd
from sklearn.metrics import confusion_matrix, classification_report, multilabel_confusion_matrix
import tensorflow as tf
from tensorflow.keras.models import load_model

from decavision.utils import data_utils
from decavision.utils import utils
from decavision.utils.training_utils import f1_score


class ModelTester:
    """
    Class to use a trained image classification model on some images. Using this
    when working with a TPU disables eager execution.

    Arguments:
        model (str): path to trained model
    """

    def __init__(self, model):
        use_tpu, use_gpu = utils.check_PU()
        if use_tpu:
            # necessary because keras generators don'T work with TPUs...
            tf.compat.v1.disable_eager_execution()
        try:
            self.model = load_model(
                model,
                custom_objects={"_f1_score": f1_score, "f1_score": f1_score})
            # efficientnets have the scaling included in them so no need to
            # rescale the images when loading
            if self.model.name[0] in ['B', 'V']:
                self.rescaling = 1
            else:
                self.rescaling = 255
            print('Model loaded correctly')
        except Exception as e:
            print('There was a problem when trying to load your model: {}'.format(e))

        self.input_shape = self.model.input_shape[1:3]

    def _load_dataset(self, path):
        """
        Load dataset into a keras generator. Images must be contained in separate
        folders for each class.

        Arguments:
            path (str): location of the dataset

        Returns:
            generator: images plus information about them (labels, paths, etc)
        """
        datagen = tf.keras.preprocessing.image.ImageDataGenerator(
            rescale=1 / self.rescaling)
        generator = datagen.flow_from_directory(directory=path,
                                                target_size=self.input_shape,
                                                shuffle=False,
                                                interpolation='bilinear',
                                                color_mode='rgb',
                                                class_mode='sparse',
                                                batch_size=1)
        return generator

    def confusion_matrix(self, path, normalize=None):
        """
        Compute and plot the confusion matrix resulting from predictions on a dataset of images.
        Images must be located in separate folders for each class.

        Arguments:
            path (str): location of the images
            normalize ('true', 'pred', 'all' or None): normalizes confusion matrix over the true (rows), predicted (columns) conditions or
                all the population. If None, confusion matrix will not be normalized.
        """
        generator = self._load_dataset(path)
        cls_true = generator.classes
        labels = list(generator.class_indices.keys())
        cls_pred = self.model.predict(generator)
        cls_pred = np.argmax(cls_pred, axis=1)
        print('Labels loaded')

        # Calculate the confusion matrix.
        cm = confusion_matrix(y_true=cls_true,  # True class for test-set.
                              y_pred=cls_pred,  # Predicted class.
                              normalize=normalize)

        # Print the confusion matrix
        ax = plt.subplot()
        sn.heatmap(cm, annot=True, ax=ax)
        ax.set_xlabel('Predicted')
        ax.set_ylabel('True')
        ax.set_title('Confusion Matrix')
        ax.xaxis.set_ticklabels(labels)
        ax.yaxis.set_ticklabels(labels)

    def _plot_images(self, images, categories, cls_true, cls_pred=None, smooth=True):
        """
        Plot images along with their true and optionally predicted labels.
        Inspired by https://github.com/Hvass-Labs/TensorFlow-Tutorials/blob/master/10_Fine-Tuning.ipynb.

        Arguments:
            images (list[numpy arrays]): list of images to plot as arrays
            categories (List[str]): list of categories that model predicts
            cls_true (np.array[int]): true labels of the images
            cls_pred (np.array[int]): predicted labels of the images
            smooth (bool): smooth out images or not when plotting
        """
        assert len(images) == len(cls_true)
        num_images = len(images)

        # Create figure with sub-plots.
        if math.sqrt(num_images).is_integer():
            nrows = ncols = int(math.sqrt(num_images))
        else:
            for i in reversed(range(math.ceil(math.sqrt(num_images)))):
                if not num_images % i:
                    nrows = int(num_images / i)
                    ncols = int(i)
                    break
        fig, axes = plt.subplots(nrows, ncols)

        # Adjust vertical spacing.
        if cls_pred:
            hspace = 0.6
        else:
            hspace = 0.3
        fig.subplots_adjust(hspace=hspace, wspace=0.3)

        # Interpolation type.
        if smooth:
            interpolation = 'spline16'
        else:
            interpolation = 'nearest'

        for i, ax in enumerate(axes.flat):
            # Plot image.
            ax.imshow(images[i], interpolation=interpolation)
            # Name of the true class.
            cls_true_name = categories[cls_true[i]]
            # Show true and predicted classes.
            if cls_pred:
                # Name of the predicted class.
                cls_pred_name = categories[cls_pred[i]]
                xlabel = "True: {0}\nPred: {1}".format(
                    cls_true_name, cls_pred_name)
            else:
                xlabel = "True: {0}".format(cls_true_name)

            # Show the classes as the label on the x-axis.
            ax.set_xlabel(xlabel)
            # Remove ticks from the plot.
            ax.set_xticks([])
            ax.set_yticks([])

        # Ensure the plot is shown correctly with multiple plots
        # in a single Notebook cell.
        plt.tight_layout()
        plt.show()

    def plot_errors(self, path, num_pictures=9):
        """
        Plot images that were not classified correctly by the model. Images to test must
        be located in separate folders for each class, for example a validation dataset.

        Arguments:
            path (str): location of the images
            num_pictures (int): maximum number of errors to show
        """
        generator = self._load_dataset(path)
        cls_true = generator.classes
        image_paths = generator.filepaths
        labels = list(generator.class_indices.keys())
        cls_pred = self.model.predict(generator)
        cls_pred = np.argmax(cls_pred, axis=1)
        print('Labels loaded')

        # get all errors index
        errors = []
        for i in range(len(cls_pred)):
            if cls_pred[i] != cls_true[i]:
                errors.append(i)

        # Load images randomly picked
        num_pictures = min(num_pictures, len(errors))
        random_errors = sorted(random.sample(errors, num_pictures))

        # Plot the images we have loaded and their corresponding classes.
        self._plot_images(
            images=[data_utils.prepare_image(image_paths[i], self.input_shape)[
                0] for i in random_errors],
            categories=labels,
            cls_true=[cls_true[i] for i in random_errors],
            cls_pred=[cls_pred[i] for i in random_errors]
        )

    def classify_images(self, image_path, categories, plot=True):
        """
        Classify images located directly in a folder. Plots the images and the first three predictions.

        Arguments:
            image_path (str): location of the images
            categories (list[str]): list of potential categories that the model can return
            plot (bool): plot or not the images, if False, only results are printed
        """
        images = glob.glob(os.path.join(image_path, '*.jpg'))
        for image_path in images:
            # prepare the image
            image_tensor = data_utils.prepare_image(
                image_path, self.input_shape, self.rescaling)
            # make and decode the prediction
            result = self.model.predict(image_tensor)[0]
            # print image and top predictions
            top_pred = np.argsort(result)[::-1][:3]
            # Name of the true class.
            cls_pred_name = np.array(categories)[top_pred]
            cls_pred_perc = result[top_pred] * 100
            if plot:
                plt.imshow(image_tensor[0], interpolation='spline16')
                xlabel = 'Prediction :\n'
                for (x, y) in zip(cls_pred_name, cls_pred_perc):
                    xlabel += '{0}, {1:.2f}%\n'.format(x, y)
                plt.xlabel(xlabel)
                plt.xticks([])
                plt.yticks([])
                plt.show()
            else:
                print('\nImage: ', image_path)
                for i in range(len(top_pred)):
                    print('Prediction: {} (probability {}%)'.format(
                        cls_pred_name[i], round(cls_pred_perc[i])))

    def evaluate(self, path):
        """
        Calculate the accuracy of the model on a dataset of images. The images must be
        in separate folders for each class.

        Arguments:
            path (str): location of the dataset
        """
        generator = self._load_dataset(path)
        results = self.model.evaluate(generator)
        print('Accuracy of', results[1] * 100, '%')

    def generate_classification_report(self, path):
        """
        Computes classification report resulting from predictions on a dataset of images
        and prints the results. Images must be located in separate folders for each class. The report shows average accuracy, precision, recall and f1-scores. Precision, recall and f1-scores are also computed for each class.

        Arguments:
            path (str): location of the images
        """
        generator = self._load_dataset(path)
        cls_true = generator.classes
        labels = list(generator.class_indices.keys())
        cls_pred = self.model.predict(generator)
        cls_pred = np.argmax(cls_pred, axis=1)
        print('Labels loaded')
        # Show classification report
        print(classification_report(
            cls_true, cls_pred, target_names=labels, digits=4))


class ModelTesterMultilabel:
    """
    Class to use a trained multilabel image classification model on some images. Using this
    when working with a TPU disables eager execution.

    Arguments:
        model (str): path to trained model
        categories (list[str]): list of potential categories that the model can return
    """

    def __init__(self, model, categories):
        use_tpu, use_gpu = utils.check_PU()
        if use_tpu:
            # necessary because keras generators don'T work with TPUs...
            tf.compat.v1.disable_eager_execution()
        try:
            self.model = load_model(model, custom_objects={
                                    "_f1_score": f1_score, "f1_score": f1_score})
            # efficientnets have the scaling included in them so no need to
            # rescale the images when loading
            if self.model.name[0] in ['B', 'V']:
                self.rescaling = 1
            else:
                self.rescaling = 255
            print('Model loaded correctly')
        except Exception as e:
            print('There was a problem when trying to load your model: {}'.format(e))

        self.input_shape = self.model.input_shape[1:3]
        self.categories = categories
        print("Model name: ", self.model.name)

    def _load_dataset(self, path, json_file):
        """
        Load dataset into a keras generator. Images must be contained in single
        folder. A dataframe is required with 2 columns: original
        image name and image labels as a list separated by comma. i.e.

        filenames                |    labels
        -------------------------|-----------------
        sun_abifwwwgjnomvfda.jpg | [asphalt,clouds,natural_light,man-made,open_area,far-away_horizon,sky,barn,airfield]
        sun_aciuvbhvntkgdhhk.jpg | [grass,asphalt,natural_light,natural,man-made,open_area,far-away_horizon,sky,dirt/soil,airfield]

        Arguments:
            path (str): location of the dataset
            json_file (str): path to json file containing image ids and their associated labels

        Returns:
            generator: images plus information about them (labels, paths, etc)
        """
        data = []
        with open(json_file, 'r') as file:
            values = json.load(file)
            for f, l in values.items():
                data.append({"filenames": f, "labels": l})

        df = pd.DataFrame(data)

        datagen = tf.keras.preprocessing.image.ImageDataGenerator(rescale=1 / self.rescaling)
        generator = datagen.flow_from_dataframe(directory=path,
                                                dataframe=df,
                                                x_col="filenames",
                                                y_col="labels",
                                                target_size=self.input_shape,
                                                shuffle=False,
                                                color_mode='rgb',
                                                class_mode='categorical',
                                                classes=self.categories,
                                                batch_size=1)
        return generator

    def classify_images(self, path, json_file, threshold=0.5, plot=False, save_img=False):
        """
        Classify images located directly in a folder. Plots and saves the images with the specified threshold.

        Arguments:
            path (str): location of the images
            json_file (str): path to json file containing image ids and their associated labels
            threshold (int): threshold for prediction (default to 0.5)
            plot (bool): plot or not the images, if False, only results are printed
            save_img (bool): save classified images or not in a new folder. To save images, you will need to set plot=True, otherwise no images will be saved. 
        """
        with open(json_file, 'r') as file:
            values = json.load(file)

        images = glob.glob(os.path.join(path, '*.jpg'))
        for image_path in images:
            # prepare the image
            image_tensor = data_utils.prepare_image(image_path, self.input_shape, self.rescaling)
            # make and decode the prediction
            result = self.model.predict(image_tensor)[0]
            top_pred = result > threshold
            # Name of the true class.
            cls_pred_name = np.array(self.categories)[top_pred]
            cls_pred_perc = result[top_pred] * 100
            cls_true = values[os.path.basename(image_path)]

            if plot:
                fig, ax = plt.subplots()
                if self.model.name[0] not in ["B", "V"]:
                    ax.imshow(image_tensor[0], interpolation='nearest')
                else:
                    ax.imshow(image_tensor[0].astype('uint8'), interpolation='nearest')
                xlabel = 'Prediction :\n'
                for (x, y) in zip(cls_pred_name, cls_pred_perc):
                    xlabel += '{0}, {1:.2f}%\n'.format(x, y)
                ax.set_xlabel(xlabel)
                ax.set_xticks([])
                ax.set_yticks([])
                plt.tight_layout()
                if save_img:
                    data_utils.create_dir("classified_images")
                    fig.savefig("classified_images/" + os.path.basename(image_path))
            else:
                print('\nImage: ', image_path)
                print("True label:", cls_true)
                for i in range(len(cls_pred_perc)):
                    print('Prediction: {} (probability {}%)'.format(cls_pred_name[i], round(cls_pred_perc[i])))

    def evaluate(self, path, json_file):
        """
        Calculate the f1-score of the model on a dataset of images. The images must be
        in single folder.

        Arguments:
            path (str): location of the dataset
            json_file (str): path to json file containing image ids and their associated labels
        """

        generator = self._load_dataset(path, json_file)
        results = self.model.evaluate(generator)
        print('f1-score of', round(results[-1] * 100, 3), '%')

    def generate_metrics(self, path, json_file, threshold=0.5):
        """
        Computes classification report and confusion matrix resulting from predictions on
        a dataset of images and prints the results. Images must be located in a single folder.

        Arguments:
            path (str): location of the images
            json_file (str): path to json file containing image ids and their associated labels
            threshold (int): threshold for prediction (default to 0.5)
        """

        # getting list of true and predicted labels
        generator = self._load_dataset(path, json_file)
        cls_pred = self.model.predict(generator)
        cls_pred = cls_pred > threshold
        cls_true = np.array([generator.next()[1][0] for i in range(generator.n)])
        print('Labels & Predictions loaded for reports')

        # classification report
        print("\nClassification report")
        print(classification_report(cls_true, cls_pred, target_names=self.categories, digits=4))

        # confusion matrix
        print("\n Confusion matrix")
        for i, item in enumerate(multilabel_confusion_matrix(cls_true, cls_pred)):
            print(self.categories[i], '\n', item, '\n')
        
    def create_movie(self, classify_images, path="", threshold=0.5, json_file="", image_folder=""):
        """
        Create a movie from classified images.

        Arguments:
            path (str): location of the images
            classify_images (bool): location of the classified images, set this to False if you have the images already saved
            threshold (int): threshold for prediction (default to 0.5)
            json_file (str): path to json file containing image ids and their associated labels
            image_folder (str): if using already classified saved images, provide path to those images. 
        """

        if classify_images:
            self.classify_images(path, json_file, threshold, plot=True, save_img=True)
            image_folder = "classified_images/"
        
        video_name = 'video.avi'

        images = [img for img in os.listdir(image_folder) if img.endswith(".jpg")]
        frame = cv2.imread(os.path.join(image_folder, images[0]))
        height, width, layers = frame.shape

        video = cv2.VideoWriter(video_name, 0, 1, (width, height))

        for image in images:
            video.write(cv2.imread(os.path.join(image_folder, image)))

        cv2.destroyAllWindows()
        video.release()



import datetime
import logging
import math
import os

import dill
import skopt
import tensorflow as tf

from decavision.utils import training_utils
from decavision.utils import utils

AUTO = tf.data.experimental.AUTOTUNE


# metric for multilable classification
class ImageClassifier:
    """
    Class to train an image classification model by using transfer learning.
    A hyperparameter optimization tool is also provided. Data must be saved in tfrecords format.
    See data_preparation.generate_tfrecords to go from image data to tfrecords.

    Arguments:
        tfrecords_folder (str): location of tfrecords (can be on google storage if authenticated), saved in
            folders train and val, filenames of the form filenumber-numberofimages.tfrec
        batch_size (int): size of batches of data used for training
        transfer_model (str): pretrained model to use for transfer learning, can be one of Inception,
            Xception, Inception_Resnet, Resnet, (EfficientNet) B0, B3, B5, B7 or (EfficientnetV2) V2-S, V2-M, V2-L
        augment (boolean): Whether to augment the training data, default is True
        input_shape (tuple(int,int)): shape of the input images for the model, if not specified, recommended sizes are used for each one
        multilable (boolean): if each image is attached to multiple classes
    """

    def __init__(self,
                 tfrecords_folder,
                 batch_size=128,
                 transfer_model='Inception',
                 augment=True,
                 input_shape=None,
                 multilabel=False):

        self.tfrecords_folder = tfrecords_folder
        self.use_TPU, self.use_GPU = utils.check_PU()
        self.multilabel = multilabel
        if multilabel:
            self.metric = "f1_score"
        else:
            self.metric = 'accuracy'
        if self.use_TPU and batch_size % 8:
            print(
                'Batch size {} is not multiple of 8, required for TPU'.format(batch_size))
            batch_size = 8 * round(batch_size / 8)
            print('New batch size is {}'.format(batch_size))
        self.batch_size = batch_size
        self.transfer_model = transfer_model
        self.augment = augment

        # We expect the classes to be saved in the same folder where tfrecords are
        self.categories = utils.load_classes(tfrecords_folder)
        print('Classes ({}) :'.format(len(self.categories)))
        print(self.categories)

        train_tfrecords = tf.io.gfile.listdir(
            os.path.join(tfrecords_folder, 'train'))
        self.nb_train_shards = len(train_tfrecords)
        print('Training tfrecords = {}'.format(self.nb_train_shards))

        val_tfrecords = tf.io.gfile.listdir(
            os.path.join(tfrecords_folder, 'val'))
        self.nb_val_shards = len(val_tfrecords)
        print('Val tfrecords = {}'.format(self.nb_val_shards))

        # Expected tfrecord file name : filenumber-numberofimages.tfrec (02-2223.tfrec)
        self.nb_train_images = 0
        for train_tfrecord in train_tfrecords:
            self.nb_train_images += int(train_tfrecord.split('.')
                                        [0].split('-')[1])
        print('Training images = ' + str(self.nb_train_images))

        nb_val_images = 0
        for val_tfrecord in val_tfrecords:
            nb_val_images += int(val_tfrecord.split('.')[0].split('-')[1])
        print('Val images = ' + str(nb_val_images))

        self.training_shard_size = math.ceil(
            self.nb_train_images / self.nb_train_shards)
        print('Training shard size = {}'.format(self.training_shard_size))

        val_shard_size = math.ceil(nb_val_images / self.nb_val_shards)
        print('Val shard size = {}'.format(val_shard_size))

        print('Training batch size = ' + str(self.batch_size))
        self.steps_per_epoch = int(self.nb_train_images / self.batch_size)
        print('Training steps per epochs = ' + str(self.steps_per_epoch))

        print('Val batch size = ' + str(self.batch_size))
        self.validation_steps = int(nb_val_images / self.batch_size)
        print('Val steps per epochs = ' + str(self.validation_steps))

        input_dims = {'Inception': 299,
                      'Xception': 299,
                      'Inception Resnet': 299,
                      'B0': 224,
                      'B3': 300,
                      'B5': 456,
                      'B7': 600,
                      'V2-S': 384,
                      'V2-M': 480,
                      'V2-L': 480}
        if transfer_model in ['B0', 'B3', 'B5', 'B7', 'V2-S', 'V2-M', 'V2-L']:
            self.scale = 255.
        else:
            self.scale = 1.

        if input_shape:
            self.target_size = input_shape
        else:
            self.target_size = (input_dims.get(self.transfer_model, 224), input_dims.get(self.transfer_model, 224))

        print("Data augmentation during training: " + str(augment))

    def _get_dataset(self, is_training, nb_readers):
        """
        Extract data from tfrecords into a tf.data iterator. If data is used for training, it is shuffled each time
        the iterator is used. Strongly inspired by
        https://colab.research.google.com/github/GoogleCloudPlatform/training-data-analyst/blob/master/courses/fast-and-lean-data-science/07_Keras_Flowers_TPU_playground.ipynb#scrollTo=LtAVr-4CP1rp

        Arguments:
            is_training (bool): set if data used for training or not, determines if data is taken from train (True)
                or val (False) tfrecords
            nb_readers (int): number of different files to combine

        Returns:
            tf.data.dataset: iterable dataset with content of relevant tfrecords (images and labels)
        """

        def _read_tfrecord(example):
            """ Extract image and label from single tfrecords example."""
            features = {
                'image': tf.io.FixedLenFeature((), tf.string),
                'label': tf.io.FixedLenSequenceFeature((), tf.int64, allow_missing=True),
            }
            example = tf.io.parse_single_example(example, features)

            image = tf.image.decode_jpeg(example['image'], channels=3)
            # normalization of pixels is already done in TF EfficientNets
            if self.transfer_model not in ['B0', 'B3', 'B5', 'B7', 'V2-S', 'V2-M', 'V2-L']:
                image = tf.image.convert_image_dtype(image, dtype=tf.float32)
            feature = tf.image.resize(image, [*self.target_size])
            label = tf.one_hot(example['label'], depth=len(self.categories), on_value=1.0, off_value=0.0)
            label = tf.reduce_sum(label, 0)
            return feature, label

        def _load_dataset(filenames):
            """ Load the tfrecords files in a tf.data object."""
            buffer_size = 8 * 1024 * 1024  # 8 MiB per file
            dataset = tf.data.TFRecordDataset(
                filenames, buffer_size=buffer_size)
            return dataset

        file_pattern = os.path.join(
            self.tfrecords_folder, "train/*" if is_training else "val/*")
        dataset = tf.data.Dataset.list_files(file_pattern, shuffle=is_training)

        # Enable non-determinism only for training.
        options = tf.data.Options()
        options.experimental_deterministic = not is_training
        dataset = dataset.with_options(options)
        dataset = dataset.interleave(
            _load_dataset, nb_readers, num_parallel_calls=AUTO)
        if is_training:
            # Shuffle only for training.
            dataset = dataset.shuffle(buffer_size=math.ceil(
                self.training_shard_size * self.nb_train_shards / 4))
        dataset = dataset.repeat()
        dataset = dataset.map(_read_tfrecord, num_parallel_calls=AUTO)
        dataset = dataset.batch(
            batch_size=self.batch_size, drop_remainder=self.use_TPU)
        dataset = dataset.prefetch(AUTO)
        return dataset

    def data_augment(self, image, label):
        """
        Data augmentation pipeline which augments the data by randomly flipping, changing brightness
        and saturation for each batch during training a model.

        References:
        https://www.wouterbulten.nl/blog/tech/data-augmentation-using-tensorflow-data-dataset/#code
        https://www.tensorflow.org/tutorials/images/data_augmentation

        Returns:
            augmented images and labels
        """
        # Flip
        image = tf.image.random_flip_left_right(image)
        # Brightness and saturation
        image = tf.image.random_brightness(image, max_delta=32.0 / 255.0)
        image = tf.image.random_saturation(image, lower=0.5, upper=1.5)
        # Make sure the image is still in [0, 1] range, after augmentation
        image = tf.clip_by_value(image, 0.0, self.scale)
        return image, label

    def get_training_dataset(self):
        """
        Extract data from training tfrecords located in tfrecords_folder. Data is shuffled and augmented.

        Returns:
            tf.data.dataset: iterable dataset with content of training tfrecords (images and labels)
        """
        if self.augment is True:
            dataset = self._get_dataset(True, self.nb_train_shards)
            # Augment data
            dataset = dataset.map(self.data_augment, num_parallel_calls=AUTO)
            return dataset
        else:
            dataset = self._get_dataset(True, self.nb_train_shards)
        return dataset

    def get_validation_dataset(self):
        """ Extract data from validation tfrecords located in tfrecords_folder.

        Returns:
            tf.data.dataset: iterable dataset with content of validation tfrecords (images and labels)
        """
        return self._get_dataset(False, self.nb_val_shards)

    def _create_model(self, activation, hidden_size, dropout, l2_lambda):
        """
        Create a keras model from a pretrained model. Add a classification layer on top and
        possibly an extra Dense layer (with dropout and batchnorm)

        Arguments:
            activation (str): activation function to use in extra layer, any keras activation is possible
            hidden_size (int): number of neurons in extra layer, no layer if 0
            dropout (float): rate of dropout to use in extra layer (<1)
            l2_lambda (float): amount of L2 regularization to include in extra layer

        Returns:
            tf.keras model: pretrained model with added layers
            int: index of the layer where the last block of the pretrained model starts
            str: name of the loss function
            list(str): names of the metrics to use
        """
        print('Creating model')
        # load the pretrained model, without the classification (top) layers
        if self.transfer_model == 'Xception':
            base_model = tf.keras.applications.Xception(weights='imagenet',
                                                        include_top=False, input_shape=(*self.target_size, 3))
            base_model_last_block = 116  # last block 126, two blocks 116
        elif self.transfer_model == 'Inception_Resnet':
            base_model = tf.keras.applications.InceptionResNetV2(
                weights='imagenet', include_top=False, input_shape=(*self.target_size, 3))
            base_model_last_block = 287  # last block 630, two blocks 287
        elif self.transfer_model == 'Resnet':
            base_model = tf.keras.applications.ResNet50(weights='imagenet',
                                                        include_top=False, input_shape=(*self.target_size, 3))
            base_model_last_block = 155  # last block 165, two blocks 155
        elif self.transfer_model == 'B0':
            base_model = tf.keras.applications.EfficientNetB0(weights='imagenet', include_top=False,
                                                              input_shape=(*self.target_size, 3))
            base_model_last_block = 213  # last block 229, two blocks 213
        elif self.transfer_model == 'B3':
            base_model = tf.keras.applications.EfficientNetB3(weights='imagenet', include_top=False,
                                                              input_shape=(*self.target_size, 3))
            base_model_last_block = 354  # last block 370, two blocks 354
        elif self.transfer_model == 'B5':
            base_model = tf.keras.applications.EfficientNetB5(weights='imagenet', include_top=False,
                                                              input_shape=(*self.target_size, 3))
            base_model_last_block = 417  # last block 559, two blocks 417

        elif self.transfer_model == 'B7':
            base_model = tf.keras.applications.EfficientNetB7(weights='imagenet', include_top=False,
                                                              input_shape=(*self.target_size, 3))
            base_model_last_block = None  # all layers trainable
        elif self.transfer_model == 'V2-S':
            base_model = tf.keras.applications.EfficientNetV2S(weights='imagenet', include_top=False,
                                                               input_shape=(*self.target_size, 3))
            base_model_last_block = 448  # last block 462, two blocks 448
        elif self.transfer_model == 'V2-M':
            base_model = tf.keras.applications.EfficientNetV2M(weights='imagenet', include_top=False,
                                                               input_shape=(*self.target_size, 3))
            base_model_last_block = 659  # last block 673, two blocks 659
        elif self.transfer_model == 'V2-L':
            base_model = tf.keras.applications.EfficientNetV2L(weights='imagenet', include_top=False,
                                                               input_shape=(*self.target_size, 3))
            base_model_last_block = 925  # last block 939, two blocks 925
        else:
            base_model = tf.keras.applications.InceptionV3(weights='imagenet',
                                                           include_top=False, input_shape=(*self.target_size, 3))
            base_model_last_block = 249  # last block 280, two blocks 249

        # Set only the top layers as trainable (if we want to do fine-tuning,
        # we can train the base layers as a second step)
        base_model.trainable = False

        # Add the classification layers using Keras functional API
        x = base_model.output
        x = tf.keras.layers.GlobalAveragePooling2D()(x)
        # Hidden layer for classification
        if hidden_size == 0:
            x = tf.keras.layers.Dropout(rate=dropout)(x)
        else:
            x = tf.keras.layers.Dense(hidden_size, use_bias=False,
                                      kernel_regularizer=tf.keras.regularizers.l2(l=l2_lambda))(x)
            # scale: When the next layer is linear (also e.g. nn.relu), this can be disabled since the
            # scaling can be done by the next layer.
            x = tf.keras.layers.BatchNormalization(
                scale=activation != 'relu')(x)
            x = tf.keras.layers.Activation(activation=activation)(x)
            x = tf.keras.layers.Dropout(rate=dropout)(x)

        x = tf.keras.layers.Dense(
            len(self.categories), name='logs')(x)  # Output layer
        if self.multilabel:
            predictions = tf.keras.layers.Activation(
                'sigmoid', name='preds')(x)  # Output activation
            loss = 'binary_crossentropy'
            metrics = ["accuracy", training_utils.f1_score]
        else:
            predictions = tf.keras.layers.Activation(
                'softmax', name='preds')(x)  # Output activation
            loss = 'categorical_crossentropy'
            metrics = ["accuracy"]

        return tf.keras.Model(inputs=base_model.input, outputs=predictions, name=self.transfer_model), base_model_last_block, loss, metrics

    def fit(self, save_model=None, export_model=None, patience=0,
            epochs=5, hidden_size=1024, learning_rate=1e-3, learning_rate_fine_tuning=1e-4,
            dropout=0.5, l2_lambda=5e-4, fine_tuning=True,
            verbose=True, logs=None, activation='swish'):
        """
        Train an image classification model based on a pretrained model. A classification layer is added
        to the pretrained model, with potentially an extra combination of Dense, Dropout and Batchnorm.
        Only added layers are trained, unless there is some fine tuning, in which case a second round of
        training is done with the last block of the pretrained model unfrozen. Training can be stopped if
        no sufficient improvement in accuracy or f1-score (in case of multilabel classification).

        If one of the Efficientnet Bs is used, the model includes a layer that normalizes the pixels. This processing
        step is not included in the other models so it has to be done on the data separately.

        Arguments:
            learning_rate (float): learning rate used when training extra layers
            learning_rate_fine_tuning (float): learning rate used when fine tuning pretrained model
            epochs (int): number of epochs done when training (doubled if fine tuning)
            activation (str): activation function to use in extra layer, any keras activation is possible
            hidden_size (int): number of neurons in extra layer, no layer if 0
            save_model (str): specify a name for the trained model to save it in .h5 format
            export_model (str): specify a name for the trained model to save it in .pb format
            dropout (float): rate of dropout to use in extra layer (<1)
            verbose (bool): show details of training or not
            fine_tuning (bool): fine tune pretrained model or not
            l2_lambda (float): amount of L2 regularization to include in extra layer
            patience (int): if non zero, stop training when improvement in val accuracy is not observed for the given number of epochs. If used, best model is restored when training is stopped
            logs (str): if specified, tensorboard is used and logs are saved at this location
        """

        # use reduce learning rate and early stopping callbacks
        reduce_lr = tf.keras.callbacks.ReduceLROnPlateau(monitor='val_' + self.metric,
                                                         factor=0.1,
                                                         patience=5,
                                                         mode='max')
        callbacks = [reduce_lr]

        if logs:
            logdir = os.path.join(
                logs, datetime.datetime.now().strftime("%Y%m%d-%H%M%S"))
            print('Fit log dir : ' + logdir)
            tensorboard_callback = tf.keras.callbacks.TensorBoard(logdir)
            callbacks.append(tensorboard_callback)

        # if we want to stop training when no sufficient improvement in validation metric has been achieved
        if patience:
            early_stop = tf.keras.callbacks.EarlyStopping(monitor='val_' + self.metric,
                                                          patience=patience,
                                                          restore_best_weights=True)
            callbacks.append(early_stop)

        # compile the model and fit the model
        if self.use_TPU:
            tpu_cluster_resolver = tf.distribute.cluster_resolver.TPUClusterResolver()
            tf.config.experimental_connect_to_cluster(tpu_cluster_resolver)
            tf.tpu.experimental.initialize_tpu_system(tpu_cluster_resolver)
            strategy = tf.distribute.TPUStrategy(
                tpu_cluster_resolver)

            with strategy.scope():
                model, base_model_last_block, loss, metrics = self._create_model(
                    activation, hidden_size, dropout, l2_lambda)
                print('Compiling for TPU')
                optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate)
                model.compile(optimizer=optimizer, loss=loss, metrics=metrics)

        else:
            model, base_model_last_block, loss, metrics = self._create_model(
                activation, hidden_size, dropout, l2_lambda)
            print('Compiling for GPU') if self.use_GPU else print(
                'Compiling for CPU')
            optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate)
            model.compile(optimizer=optimizer, loss=loss, metrics=metrics)

        print('Fitting')
        history = model.fit(self.get_training_dataset(), steps_per_epoch=self.steps_per_epoch, epochs=epochs,
                            validation_data=self.get_validation_dataset(), validation_steps=self.validation_steps,
                            verbose=verbose, callbacks=callbacks)

        # Fine-tune the model, if we wish so
        if fine_tuning and not model.stop_training:
            print('===========')
            print('Fine-tuning')
            print('===========')

            fine_tune_epochs = epochs
            total_epochs = epochs + fine_tune_epochs

            print('Unfreezing last block of layers from the base model')
            for layer in model.layers[:base_model_last_block]:
                layer.trainable = False
            # don't unfreeze batchnorm (see https://keras.io/examples/vision/image_classification_efficientnet_fine_tuning/)
            for layer in model.layers[base_model_last_block:]:
                if not isinstance(layer, tf.keras.layers.BatchNormalization):
                    layer.trainable = True

            # Fit the model
            # we need to recompile the model for these modifications to take effect with a low learning rate
            print('Recompiling model')
            optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate_fine_tuning)
            model.compile(optimizer=optimizer, loss=loss, metrics=metrics)

            print('Fine tunning')
            history = model.fit(self.get_training_dataset(), steps_per_epoch=self.steps_per_epoch, epochs=total_epochs,
                                validation_data=self.get_validation_dataset(), validation_steps=self.validation_steps,
                                verbose=verbose, callbacks=callbacks, initial_epoch=epochs)

        # Evaluate the model, just to be sure
        self.fitness = history.history['val_' + self.metric][-1]
        self.model = model
        del history
        del model
        # Save the model
        if save_model:
            self.model.save(save_model + '.h5')
            print('Model saved')
        if export_model:
            self.model.save(export_model)
            print('Model exported')

    def hyperparameter_optimization(self, num_iterations=20, n_random_starts=10, patience=0, save_results=False):
        """
        Try different combinations of hyperparameters to find the best model possible. Start by trying random
        combinations and after some time learn from th previous tries. Scikit-optimize checkoint is saved
        at each step in the working directory. If checkpoint present in working directory, optimization starts
        back from where it left off. Logs of all tries are also saved in working directory. Hyperparameters that
        are varied are epochs, hidden_size, learning_rate, learning_rate_fine_tuning, fine_tuning, dropout and
        l2_lambda. Possible to save best combination at the end of the optimization.

        Arguments:
            n_random_starts (int): number of random combinations of hyperparameters first tried
            num_iterations (int): total number of hyperparameter combinations to try (aim for a 1:1 to 2:1 ratio
                num_iterations/n_random_starts)
            patience (int): if non zero, stop training when improvement in val accuracy is not observed for the given number of epochs. If used, best model is restored when training is stopped
            save_results (bool): decide to save optimal hyperparameters in hyperparameters_dimensions.pickle when done
        """
        # initialize logging
        logging.basicConfig(
            level=logging.INFO,
            format="%(asctime)s [%(levelname)s] %(message)s",
            handlers=[
                logging.FileHandler("hyperparameters.log"),
                logging.StreamHandler()
            ]
        )
        logging.getLogger('googleapiclient.discovery_cache').setLevel(logging.ERROR)
        # declare the hyperparameters search space
        dim_epochs = skopt.space.Integer(low=1, high=6, name='epochs')
        dim_hidden_size = skopt.space.Integer(
            low=512, high=2048, name='hidden_size')
        dim_learning_rate = skopt.space.Real(low=1e-6, high=1e-2, prior='log-uniform',
                                             name='learning_rate')
        dim_learning_rate_fine_tuning = skopt.space.Real(low=1e-6, high=1e-2, prior='log-uniform',
                                                         name='learning_rate_fine_tuning')
        dim_dropout = skopt.space.Real(low=0, high=0.9, name='dropout')
        dim_l2_lambda = skopt.space.Real(low=1e-6, high=1e-2, prior='log-uniform',
                                         name='l2_lambda')
        dim_fine_tuning = skopt.space.Categorical(categories=[True, False],
                                                  name='fine_tuning')

        dimensions = [dim_epochs,
                      dim_hidden_size,
                      dim_learning_rate,
                      dim_learning_rate_fine_tuning,
                      dim_dropout,
                      dim_l2_lambda,
                      dim_fine_tuning]

        # read default parameters from last optimization
        try:
            res = skopt.load('checkpoint.pkl')
            x0 = res.x_iters
            y0 = res.func_vals
            start_from_checkpoint = True
            print('Parameters of previous optimization loaded!')
        except:
            # fall back default values
            default_parameters = [2, 1024, 5e-4, 6e-4, 0.9, 1e-3, True]
            start_from_checkpoint = False

        checkpoint_saver = skopt.callbacks.CheckpointSaver(
            'checkpoint.pkl', store_objective=False)
        checkpoint_downloader = training_utils.CheckpointDownloader(
            'checkpoint.pkl')
        verbose = skopt.callbacks.VerboseCallback(n_total=num_iterations)

        @skopt.utils.use_named_args(dimensions=dimensions)
        def fitness(epochs, hidden_size, learning_rate, learning_rate_fine_tuning, dropout, l2_lambda, fine_tuning):
            """
            Function to be minimized by the optimization. Trains a model using the fit method and the given
            hyperparameters to return the final (negative) value of the validation accuracy.
            """
            # print the hyper-parameters
            logging.info('Fitnessing hyper-parameters')
            logging.info(f'epochs:{epochs}')
            logging.info(f'hidden_size:{hidden_size}')
            logging.info(f'learning rate:{learning_rate}')
            logging.info(
                f'learning rate fine tuning:{learning_rate_fine_tuning}')
            logging.info(f'dropout:{dropout}')
            logging.info(f'l2_lambda:{l2_lambda}')
            logging.info(f'fine_tuning:{fine_tuning}')

            # fit the model
            self.fit(epochs=epochs, hidden_size=hidden_size, learning_rate=learning_rate,
                     learning_rate_fine_tuning=learning_rate_fine_tuning,
                     dropout=dropout, l2_lambda=l2_lambda, fine_tuning=fine_tuning,
                     patience=patience)

            # extract fitness
            fitness = self.fitness

            logging.info(f'CALCULATED FITNESS OF:{fitness}')

            del self.model
            tf.keras.backend.clear_session()
            return -fitness

        # optimization
        if start_from_checkpoint:
            print('Continuous fitness')
            search_result = skopt.forest_minimize(func=fitness,
                                                  dimensions=dimensions,
                                                  x0=x0,  # already examined values for x
                                                  y0=y0,  # observed values for x0
                                                  # Expected Improvement.
                                                  acq_func='EI',
                                                  n_calls=num_iterations,
                                                  n_random_starts=n_random_starts,
                                                  callback=[checkpoint_saver, checkpoint_downloader, verbose])
        else:
            print('New fitness')
            search_result = skopt.forest_minimize(func=fitness,
                                                  dimensions=dimensions,
                                                  # Expected Improvement.
                                                  acq_func='EI',
                                                  n_calls=num_iterations,
                                                  n_random_starts=n_random_starts,
                                                  x0=default_parameters,
                                                  callback=[checkpoint_saver, checkpoint_downloader, verbose])

        if save_results:
            with open('hyperparameters_dimensions.pickle', 'wb') as f:
                dill.dump(search_result.x, f)
            print("Hyperparameter results saved!")

        # build results dictionary
        results_dict = {dimensions[i].name: search_result.x[i]
                        for i in range(len(dimensions))}
        logging.info(
            'Optimal fitness value of:{}'.format(-float(search_result.fun)))
        logging.info('Optimal hyperparameters:{}'.format(results_dict))


if __name__ == '__main__':
    classifier = ImageClassifier()
